Unit TTextV;

   { Written by Mark Cole, PascalNet 115:4401/1 (2:441/286@Fideonet)
     Version 1.00 March 1995. }

INTERFACE

USES Dos,
     Objects, Drivers, Memory, Views, Dialogs, MsgBox, TextView, App;


Const
  { Max lines allowed in TTextInterior (minus one as is zero based) }
  TextInteriorMaxLines = 4095;

  {TTextView Styles}
  ttvFullHeight  = $0001;  { Text fills dialog vertically }
  ttvFullWidth   = $0002;  {   "    "     "    horizontally}
  ttvPrint       = $0004;  { allow user to print (implied by ttvPrintButton)}
  ttvSearch      = $0008;  { allow user to search (implied by ttvSearchButton) }
  ttvPrintButton = $0010;  { add a [Print] button (not if Full Height }
  ttvOkButton    = $0020;  { ditto [Ok]  }
  ttvCancelButton= $0040;  { guess! }
  ttvSearchButton= $0080;  { add [Search] and [Again] buttons }
  ttvUnFramed    = $0100;  { don't drame the dialog}
  ttvHideVScroll = $0200;  { hide vertical scroll bar in full sceen view }
  ttvHideHScroll = $0400;

  { Couple of standard settings }
  TextViewNormal = ttvSearchButton or ttvOkButton or ttvPrintButton;

  TextViewFull   = ttvUnframed or ttvPrint or ttvSearch
                   or ttvFullHeight or ttvFullWidth;


  { These three are only used here, alter if any conflict with your app}
  cmSearch = 100;
  cmAgain  = 101;
  cmPrint  = 110;

  { Hint/Help context for viewer }
  hcTextViewer = $FFFE; { Or whatever suits your application }

  TagCode = #255;       { Marks a tagged line in TTerminal, should be
                          something that will not normally occur at s[1]
                          of any text line }

TYPE

  { TTextInterior is an improved TTerminal that allows the interior of
    a window (Typically a TTextViewer) to be treated as a virtual screen
    with methods to position the write cursor and to WriteAt(x,y). As
    with TTerminal a text file is also available for use with Write/Writeln
    and this always writes at the current write cursor position.

    A mouse double-click "tags" a line (but only TTextViewer implements
    methods to retrieve tagged lines later). Tagged lines are highlighted.

    Var "HighlitLine" may be set to highlight a line (TTextViewer does
    this to highlight a line found by its text search).
    }

  TTextInteriorArray = Array[0..TextInteriorMaxLines] of PString;
  PTextInterior = ^TTextInterior;
  TTextInterior = object(TTextDevice)
    CurX,
    CurY,
    HighLine,
    HighCol,
    MaxLines: integer;
    Lines: ^TTextInteriorArray;
    LockFlag: byte;
    HighlitLine: integer;
    Constructor Init(VAR Bounds:TRect; AHScrollBar, AVScrollBar: PScrollBar;
                             AMaxLines: integer);
    Destructor Done;                                                  Virtual;
    Procedure Draw;                                                   Virtual;
    Function  GetPalette: PPalette;                                   Virtual;
    Procedure HandleEvent(VAR Event: TEvent);                         Virtual;
    Procedure StrWrite(var S: TextBuf; Count: Byte);                  Virtual;
    Procedure WriteAt(x,y: integer; S: String);
    Procedure WriteLNAt(x,y: integer; S: String);
    Procedure GotoXY(x,y: byte);
    Function WhereX: word;
    Function WhereY: word;
  end;

  PTextViewer = ^TTextViewer;
  TTextViewer = object(TDialog)
    Terminal:   PTextInterior;
    Style:      word;     { Copy of init's AStyle }
    PrintFile:  text;
    SearchPos:  integer; { last search line }
    TaggedLine: integer; { used by First/NextTaggedLine}
    constructor Init(var Bounds: TRect;
                         AStyle: word;
                         ATitle: TTitleStr;
                         AString: String;
                         Var AFile: Text);
    Destructor Done;                                                  Virtual;
    Procedure WriteAt(x,y: integer; St: String);                      Virtual;
    Procedure WriteLNAt(x,y: integer; St: String);                    Virtual;
    Procedure HandleEvent(VAR Event: TEvent);                         Virtual;
    Procedure PrintLine(s: string);                                   Virtual;
    Procedure PrintView;                                              Virtual;
    procedure Search(GetValues: boolean);                             Virtual;
    procedure HighlightLine(Line: integer);                           Virtual;
    Function  FirstTaggedLine: string;                                Virtual;
    Function  NextTaggedLine: string;                                 Virtual;
    Procedure InitFrame;                                              Virtual;
    Procedure Draw;                                                   Virtual;
    Procedure CalcBounds(var Bounds: TRect; Delta: TPoint); virtual;
  end;


IMPLEMENTATION

CONST

   { These search criterion could be local to TTextViewer but my original
     app wanted the same settings to apply globally across all views }
   LastSearchString: string = '';
   IgnoreCase:       boolean = true;
   FromTop:          boolean = false;

   { Palette for TTextInterior. }
   CTextInterior = #25#27#26; { Maps to TDialogs palette and uses entry 26
                                (ListViewer Normal) for text in the text
                                view and entry 27 (ListViewr Selected)
                                for highlit text (eg. text found in
                                a search).

                                Entry 25 (History Icons) is for "tagged"
                                lines (was supposed to be 30, InfoPane) but
                                for some reason attempts to use 30 always
                                return the value of 29 (List divider).
                                It MUST ust be me but I'm damned if I can
                                where where - it all looks correct!!&&^^!

                                This palette can be altered as you wish.
                              }


VAR GlobalStyle: word;     {Global copy of TextViewer style flag (see Init)}


{ Some support routines, replace where appropriate with those in your
  own library }

Procedure UpCaseStr(var s: string); assembler;
asm
  mov   dx,ds      { save DS in DX }
  lds   si,s       { DS:SI is address of s }
  mov   cl,[si]    { CL := length(s) }
  xor   ch,ch      { CH := 0, means CX=CL }
  jcxz  @done      { finished if CX = 0 }
@loop:
  inc   si         { si -> next character in string }
  mov   al,[si]    { al := s[si] }
  cmp   al,'a'     { al < 'a' }
  jb    @nochange
  cmp   al,'z'     { al > 'z' }
  ja    @nochange
  and   al,not 32
  mov   [si],al    { s[si] : =al }
@nochange:
  loop  @loop      { dec(cx), repeat until cx=0 }
@done:
  mov   ds,dx      { restore DS (vital) }
end;

Function UpCaseStrF(S: string): string;
var i: word;
begin
  if length(s) > 0 then for i:=1 to length(S) do if s[i] in ['a'..'z']
    then s[i]:=UpCase(s[i]);
  UpcaseStrF:=s;
end;

  { A messagebox interface that simply clears a pending event }
Procedure Msg(Msg: string; MsgParms: pointer; MsgFlags: word);
var E: TEvent;
begin
  Application^.GetEvent(E); { clear a pending event }
  MessageBox(Msg,MsgParms,MsgFlags);
end;

Procedure LowMemMsg;
begin
  Msg('Not enough memory available.',nil,mfWarning+mfOkButton);
end;

Function LowMem: boolean;
begin
  if LowMemory or (MaxAvail < 2048) then
   begin
     LowMemMsg;
     LowMem := true;
   end
  else LowMem := false;
end;

Function FileError(FileName: PathStr): boolean;
var i: integer;
    s: string[40];
begin
  i := ioresult;
  if i > 0 then
   begin
     Msg(^C'File error on '+FileName,Nil,mfError+mfOkButton);
     FileError := true;
   end else FileError := false;
end;


{ These can be made local to TTextViewer but are global here as several
  views use them }
Function GetSearchValues: boolean;
var
  SrchRec : record
    SearchFor : String[60];
    CaseSensitive:   Word;
    FromTop:    Word;
  end;

  Dlg : PDialog;
  R : TRect;
  Control : PView;

begin
  GetSearchValues := False;
  if LowMemory then exit;
  R.Assign(10, 5, 52, 15);
  New(Dlg, Init(R, ''));
  {
  Dlg^.Palette := ColorSets[Setup.ColourSet][ocSearch];
  Dlg^.HelpCtx := hcSearch;
  }
  R.Assign(14, 2, 39, 3);
  Control := New(PInputLine, Init(R, 60));
  Dlg^.Insert(Control);

  R.Assign(2, 2, 14, 3);
  Dlg^.Insert(New(PLabel, Init(R, 'Search For:', Control)));
  (* *)
  R.Assign(3, 4, 25, 5);
  Control := New(PCheckboxes, Init(R,
  NewSItem('~C~ase Sensitive',nil)));
  Dlg^.Insert(Control);

  R.Assign(3, 5, 25, 6);
  Control := New(PCheckboxes, Init(R,
  NewSItem('Search From ~T~op', Nil)));
  Dlg^.Insert(Control);
  (* *)
  R.Assign(9, 7, 19, 9);
  Control := New(PButton, Init(R, 'OK', cmOk, bfNormal+bfDefault));
  Dlg^.Insert(Control);

  R.Assign(24, 7, 34, 9);
  Control := New(PButton, Init(R, 'Cancel', cmCancel, bfNormal));
  Dlg^.Insert(Control);

  Dlg^.SelectNext(False);
  SrchRec.SearchFor := LastSearchString;
  if IgnoreCase then SrchRec.CaseSensitive := 0 else SrchRec.CaseSensitive := 1;
  if FromTop then SrchRec.FromTop := 1 else SrchRec.FromTop := 0;
  if Application^.ExecuteDialog(Dlg,@SrchRec) = cmOk then
   begin
     LastSearchString := SrchRec.SearchFor;
     IgnoreCase := SrchRec.CaseSensitive = 0;
     FromTop := SrchRec.FromTop > 0;
     GetSearchValues := LastSearchString <> '';
   end;
end;


{_______________ TTextInterior ___________________}

Constructor TTextInterior.Init;
begin
  TTextDevice.Init(Bounds, AHScrollBar, AVScrollBar);
  GrowMode := gfGrowHiX + gfGrowHiY;
  EventMask := EventMask or evMouseUp or evMouseauto;
  MaxLines := AMaxLines;
  if MaxLines > TextInteriorMaxLines then MaxLines := TextInteriorMAxLines;
  if (MaxAvail < MaxLines * 4) then
   begin
     LowMemMsg;
     Fail;
   end;
  GetMem(Lines, MaxLines * 4);
  FillChar(Lines^,MaxLines * 4, 0);
  HighLine := 0;
  HighCol := 0;
  CurX := 0;
  CurY := 0;
  HighLitLine := -1;
  SetLimit(0,1);
  SetCursor(0,0);
end;

Function TTextInterior.GetPalette;
const
  P: String[Length(CTextInterior)] = CTextInterior;
begin
  GetPalette := @P;
end;

Procedure TTextInterior.HandleEvent;
var a:byte;
    R: TRect;
    Mouse: TPoint;

  Procedure MouseScroll;
  var BoxL,BoxR,BoxU,BoxD: TRect;

    Procedure SUp;
    begin
      if Delta.Y > 0 then
       begin
         ScrollTo(Delta.X,Delta.Y-1);
       end;
    end;

    Procedure SDn;
    begin
      if (Delta.Y + Size.Y + 2) < HighLine then
       begin
         ScrollTo(Delta.X,Delta.Y+1);
       end;
    end;

    Procedure SLf;
    begin
      if Delta.X > 0 then
       begin
         ScrollTo(Delta.X-1,Delta.Y);
       end;
    end;

    Procedure SRt;
    begin
      if (Delta.X + Size.X - 1) < HighCol then
       begin
         ScrollTo(Delta.X+1,Delta.Y);
       end;
    end;

  begin
    BoxL.Assign(0,0,10,Size.Y);
    BoxR.Assign(Size.X-10,0,Size.X,Size.Y);
    BoxU.Assign(0,0,Size.X,pred(Size.Y div 2));
    BoxD.Assign(0,succ(Size.Y div 2),Size.X,Size.y);
    if BoxU.Contains(Mouse) then SUp else if BoxD.Contains(Mouse) then SDn;
    if BoxL.Contains(Mouse) then SLf else if BoxR.Contains(Mouse) then SRt;
    ClearEvent(Event);
  end;

  Procedure MouseMark;
  var l: longint;
      St: string;
  begin
    l := Delta.Y + Mouse.Y;
    if Lines^[L] <> nil then
     begin
       St := Lines^[L]^;
       if St[1] = TagCode then System.Delete(St,1,1) else St := TagCode + St;
       DisposeStr(Lines^[L]);
       if length(St) > 0 then Lines^[L] := NewStr(St) else Lines^[L] := NIL;
       DrawView;
     end;
    ClearEvent(Event);
  end;

begin
  inherited HandleEvent(Event);
  if (Event.What and evMouse > 0) and MouseInView(Event.Where) then
   begin  { Mouse event in our area }
     MakeLocal(Event.Where,Mouse);
     if (Event.Buttons > 0) and (Event.What = evMouseAuto)
       or ((Event.What = evMouseDown) and (Event.Buttons = mbRightButton))
       then MouseScroll
     else if Event.Double and (Event.Buttons = mbLeftButton) then MouseMark;
   end;
end;

Procedure TTextInterior.Draw;
var i,j,k,
    BottomLine: integer;
    St:         string;

begin
  BottomLine := Size.Y + Delta.Y;
  if BottomLine > Limit.Y then BottomLine := Limit.Y;
  For i := Delta.Y to BottomLine do
   begin
     if Lines^[I] <> nil then
      begin
        St := Lines^[I]^;
        if St[1] = TagCode then k := 1
         else if I = HighlitLine then k := 2
          else k := 3;
        if K = 1 then
         begin
           delete(St,1,1); { "tagged" }
         end;
        if Delta.X < length(St) then St := Copy(St,Delta.X+1,Size.X)
         else St := '';
        WriteStr(0,I-Delta.Y,St,k);
      end
     else St := '';
     if Length(St) < Size.X then WriteChar(Length(St),I-Delta.Y,' ',3,Size.X-Length(St));
   end;
  inc(i);
  if i-Delta.Y < Size.Y then
   for i := i-Delta.Y to Size.Y do WriteChar(0,I,' ',3,Size.X);
end;

Procedure TTextInterior.WriteAt;
var St: string;
begin
  if x < 0 then x := CurX else if x > 255 then x := 255;
  inc(x); { 1 based in strings }
  if y < 0 then y := CurY else if y >= MaxLines then y := pred(MaxLines);
  fillchar(St[1],255,#32);

  if Lines^[Y] = nil then
   begin
     St[0] := Char(x-1);
     St := St + s;
     Lines^[Y] := NewStr(St);
   end
  else
   begin
     St := Lines^[Y]^;
     if length(s) + x > 255 then s[0] := chr(255-x);
     move(s[1],St[X],Length(S));
     if length(St) < pred(X + Length(s)) then St[0] := chr(pred(X + Length(s)));
     if Length(St) <> Length(Lines^[Y]^) then
      begin
        DisposeStr(Lines^[Y]);
        Lines^[Y] := NewStr(St);
      end
     else Lines^[Y]^ := St;
   end;
  if HighLine < Y then HighLine := y;
  if HighCol < Length(St)-1 then HighCol := Length(St)-1;
  SetLimit(HighCol+1,HighLine+1);
  if LockFlag = 0 then
   begin
     ScrollTo(0,CurY+1);
     DrawView;
   end;
  CurY := y;
  CurX := pred(x + Length(s));
end;

Procedure TTextInterior.WriteLNAt;
begin
  WriteAt(x,y,S);
  CurX := 0;
  if CurY < pred(MaxLines) then inc(CurY);
End;

Procedure TTextInterior.GotoXY;
begin
  if (x < 256) and (x >- 1) then CurX := x;
  if (y <= pred(MaxLines)) and (y > -1) then CurY := y;
end;

Function TTextInterior.WhereX; begin WhereX := CurX; End;
Function TTextInterior.WhereY; begin WhereY := CurY; End;

Procedure TTextInterior.StrWrite;
var St: string;
    i:  integer;
begin
  if Count = 0 then exit;
  St := '';
  inc(LockFlag);
  for i := 0 to Count - 1 do
   begin
     case S[i] of
           #13: begin
                  if St <> '' then WriteAt(CurX,CurY,St);
                  CurX := 0;
                  St := '';
                end;
           #10: begin
                  if St <> '' then WriteAt(CurX,CurY,St);
                  St := '';
                  if CurY < pred(MaxLines) then
                   begin
                     inc(CurY);
                     if CurY > HighLine then inc(HighLine);
                   end;
                end;
           else St := St + S[i];
     end; {case}
  end; {for}
  if St <> '' then WriteAt(CurX,CurY,St);
  dec(LockFlag);
  ScrollTo(0,CurY+1);
  DrawView;
end;

Destructor TTextInterior.Done;
var i: integer;
begin
  for i := 0 to HighLine do if Lines^[i] <> nil then DisposeStr(Lines^[i]);
  freemem(Lines,MaxLines * 4);
  Inherited Done;
End;


{_________________________ TTextViewer __________________________}
Constructor TTextViewer.Init;
var
  R: TRect;
  HScrollBar,
  VScrollBar: PScrollBar;
  Control:    PView;
  TLines:     byte;
  i:          integer;
  xa,ya:      byte;        { x and y adjust for interior }
  yb:         byte;        { button lines adjust for bottom }
  OkType:     word;        { ok button, bfdefault or bfnormal }

begin
    { Somewhat scruffy, we can't use the "style" var in InitFrame because
      the call to TObject.Init initializes all fields and we loose the value
      so in order for InitFrame to know whether we want a frame or not we
      must temporarily store the "Style" value in a global variable.
      Very un-Oop'ish :-}

  GlobalStyle := AStyle;

  inherited Init(Bounds, ATitle);
  Style := AStyle;
  if LowMemory then fail;
  SearchPos := -1;

  { How many lines of static text in AString }
  TLInes := 0;
  if AString <> '' then
   begin
     TLines := 1;
     for i := 1 to length(AString) do if AString[i] = #13 then inc(TLines);
     Bounds.Assign(1,2,Size.X-1,2+TLines);
     Control := new(PStaticText,Init(Bounds,AString));
     Insert(Control);
     inc(TLines);      { Leave gap before text pane }
   end;

  { Bounds for text view (TCTerminal) }
  if AStyle and ttvUnframed > 0 then xa := 1  { no frame }
   else if AStyle and ttvFullWidth > 0 then xa := 2
   else xa := 4;
  if AStyle and ttvUnframed > 0 then  Ya := 0  { no frame }
   else if (AStyle and ttvFullHeight > 0) and (AString='') then ya := 1
   else ya := 2;
  if (AStyle and ttvUnframed > 0) then yb := 1
   else if (AStyle and (ttvPrintButton or ttvOkButton or ttvCancelButton or ttvSearchButton) > 0)
     then yb := 5 else yb := 2;

  Bounds.Assign(xa-1,ya+Tlines,Size.X-xa,Size.Y-yb);
  R.Assign(Bounds.A.X, Bounds.B.Y, Bounds.B.X+1, Bounds.B.Y+1);
  HScrollBar := new(PScrollBar,Init(r));
  Insert(HScrollBar);
  HScrollBar^.Options := HScrollBar^.Options or ofPostProcess;

  R.Assign(Bounds.B.X,Bounds.A.Y,Bounds.B.X+1,Bounds.B.Y);
  VScrollBar := new(PScrollBar,Init(r));
  Insert(VScrollBar);
  VScrollBar^.Options := VScrollBar^.Options or ofPostProcess;

  if AStyle and ttvHideVScroll > 0 then inc(Bounds.B.X);
  if AStyle and ttvHideHScroll > 0 then inc(Bounds.B.Y);
  New(Terminal, Init(Bounds, HScrollBar, VScrollBar, TextInteriorMaxLines));
  if Application^.ValidView(Terminal) <> nil then
  begin
    AssignDevice(AFile, Terminal);
    Rewrite(AFile);
    Insert(Terminal);
  end
  else fail;

  R.Assign(2,Size.Y-3,12,Size.Y-1);
  if AStyle and ttvOkButton > 0 then
   begin
     if AStyle and ttvCancelButton > 0 then OkType := bfNormal
      else OkType := bfDefault; { default if no cancel button }
     Control := New(PButton, Init(R, '~O~k', cmOk, OkType));
     Control^.GrowMode := gfGrowAll;
     Insert(Control);
     inc(R.A.X,12); inc(R.B.X,12);
   end;

  if AStyle and ttvCancelButton > 0 then
   begin
     Control := New(PButton, Init(R, '~C~ancel', cmCancel, bfDefault));
     Control^.GrowMode := gfGrowAll;
     Insert(Control);
     inc(R.A.X,12); inc(R.B.X,12);
   end;

  if AStyle and ttvPrintButton > 0 then
   begin
     Control := New(PButton, Init(R, '~P~rint', cmPrint, bfNormal));
     Control^.GrowMode := gfGrowAll;
     Insert(Control);
     inc(R.A.X,12); inc(R.B.X,12);
     Style := Style or ttvPrint;
   end;

  if AStyle and ttvSearchButton > 0 then
   begin
     Control := New(PButton, Init(R, '~S~earch', cmSearch, bfNormal));
     Control^.GrowMode := gfGrowAll;
     Insert(Control);
     inc(R.A.X,12); inc(R.B.X,12);
     Control := New(PButton, Init(R, '~A~gain', cmAgain, bfNormal));
     Control^.GrowMode := gfGrowAll;
     Insert(Control);
     inc(R.A.X,12); inc(R.B.X,12);
     Style := Style or ttvSearch;
   end;

  SelectNext(false);
  SetState(sfVisible,false);
end;


Destructor TTextViewer.Done;
begin
  Inherited Done;
end;

Procedure TTextViewer.PrintLine(s: string);
const LeftMargin = 72;
var i: integer;
begin
  Writeln(PrintFile,S);
end;

Procedure TTextViewer.PrintView;
const S: string = 'PRN';
var i: integer;

begin
  if InputBox('Print Area','Print File',s,64) = cmOK then
   begin
     assign(PrintFile,s);
     rewrite(PrintFile);
     if FileError(s) then exit;
     i := 0;
     with Terminal^ do
      repeat
       if Lines^[i] <> nil then PrintLine(Lines^[i]^) else PrintLine('');
       inc(i);
      until (i > HighLine) or FileError(s);
     System.Close(PrintFile);
  end;
end;

procedure TTextViewer.HighlightLine(Line: integer);
var l: integer;
begin
  Terminal^.HighLitLine := Line;
  { Center found line on screen if poss }
  if Line > Size.Y div 2 then l := Line - (Size.Y div 2)
   else l := 0;
  Terminal^.ScrollTo(0,l);
end;

procedure TTextViewer.Search(GetValues: boolean);

var
  Fp: byte;
  l:  integer;
  s,s1: string;

begin
  if GetValues then
   begin
    if not GetSearchValues then exit;
    if FromTop then SearchPos := -1 else SearchPos := Terminal^.Delta.Y;
   end;
  if SearchPos < Terminal^.HighLine then { not at end after last repeat }
   begin
     Terminal^.HighlitLine := -1;
     s := LastSearchString;
     if IgnoreCase then UpcaseStr(s);
     Fp := 0;
     with Terminal^ do repeat
       inc(SearchPos);
       if Lines^[SearchPos] <> nil then
        begin
          if IgnoreCase then s1 := UpcaseStrF(Lines^[SearchPos]^)
           else s1 := Lines^[SearchPos]^;
          Fp := pos(s,s1);
        end;
     until (SearchPos = HighLine) or (Fp > 0);
     if Fp > 0 then
      begin
        HighlightLine(SearchPos);
        Terminal^.DrawView;
      end
     else Msg(#3'Not Found',nil,mfInformation+MfOkButton);
   end
  else Msg(#3'Not Found',nil,mfInformation+MfOkButton);
end;

Procedure TTextViewer.CalcBounds(var Bounds: TRect; Delta: TPoint);
var R: TRect;
begin
  Owner^.GetExtent(Bounds);
 { inherited CalcBounds(Bounds,Delta);}
end;

Procedure TTextViewer.HandleEvent;
  { The cmOK/cmCancel (and other standard commands?) only gets handled here
    if we are NOT modal. When modal (executed via ExecuteDialog) they get
    handled by TDialog which will end the modal state.

    cmPrint is always done here }
var a: byte;
    R: TRect;

begin
  inherited HandleEvent(Event);
  if Event.What = evCommand then
  begin
    case Event.Command of
      cmQuit,           { TDialog.HandleEvent does not handle cmQuit }
      cmOk,             { which is what we get from ESC on the status line }
      cmCancel: begin
                  EndModal(Event.Command);
                  ClearEvent(Event);
                end;
      cmSearch: if Style and ttvSearch > 0 then
                 begin
                  Search(True);
                 end;
      cmAgain:  if Style and ttvSearch > 0 then
                 begin
                  Search(LastSearchString = '');
                 end;
      cmPrint:  if Style and ttvPrint > 0 then
                 begin
                   PrintView;
                   ClearEvent(Event);
                 end;
    end;
  end;
end;

Procedure TTextViewer.WriteAt;
begin
  Terminal^.WriteAt(x,y,St);
End;

Procedure TTextViewer.WriteLNAt;
begin
  Terminal^.WriteLNAt(x,y,St);
End;

Procedure TTextViewer.InitFrame;
begin
  if GlobalStyle and ttvUnframed = 0 then Inherited InitFrame else Frame := nil;
end;

Procedure TTextViewer.Draw;
begin
  Palette := dpBlueDialog;
  TaggedLine := 0;
  Inherited Draw;
end;

Function TTextViewer.FirstTaggedLine;
begin
  TaggedLine := 0;
  FirstTaggedLine := NextTaggedLine;
end;

Function TTextViewer.NextTaggedLine;
begin
  with Terminal^ do
   begin
     while (TaggedLine <= HighLine)
       and (((Lines^[TaggedLine] <> nil)
       and (Lines^[TaggedLine]^[1] <> TagCode))
        or (Lines^[TaggedLine] = nil)) do inc(TaggedLine);

     if TaggedLine <= HighLine then
      begin
        NextTaggedLine := Copy(Lines^[TaggedLine]^,2,255);
        Inc(TaggedLine);
      end
     else NexttaggedLine := '';
  end;
end;

END.
